#include "node/components/context.h"
#include "node/components/control.h"
#include "node/components/telemetry.h"
#include "node/load.h"
#include "node/servos.h"
#include "shared/protocol.h"

#include <emlabcpp/zip.h>
#include <schpin_lib/json.h>
#include <schpin_msgs/BlobCmd.h>
#include <schpin_msgs/LuaCall.h>
#include <schpin_msgs/LuaDef.h>
#include <std_msgs/String.h>

using namespace schpin;

class node
{

        node_hardware_config config_;
        ros::NodeHandle      handle_;

        ros::Publisher  control_pub_;
        ros::Subscriber control_sub_;

        lua_vm_interactive lua_vm_;
        ros::ServiceServer lua_def_;
        ros::ServiceServer lua_call_;

        context_component   context_comp_;
        control_component   ctl_comp_;
        telemetry_component telem_comp_;

        ctl_id msg_id_ = 0;

        struct feedback_handler
        {
                node& node_;

                ctl_id operator()( machine_feedback_var var )
                {
                        return var << match(
                                   [&]( node_fw_group_message msg ) -> ctl_id {
                                           return node_.transmit( msg );
                                   },
                                   [&]( set_body_pose_event e ) -> ctl_id {
                                           node_.force_body_pose( e.i_var );
                                           return ctl_id{ 0 };
                                   },
                                   [&]( servo_config_event e ) -> ctl_id {
                                           node_.set_servo_config( e.id, e.json );
                                           return ctl_id{ 0 };
                                   } );
                }
        };

      public:
        node( const node& ) = delete;
        node( node&& )      = delete;
        node& operator=( const node& ) = delete;
        node& operator=( node&& ) = delete;

        node( const ros::NodeHandle& handle, node_hardware_config c, robot_model rob )
          : config_( std::move( c ) )
          , handle_( handle )
          , control_pub_( handle_.advertise< schpin_msgs::BlobCmd >( "/control/control", 200 ) )
          , control_sub_( handle_.subscribe( "/control/feedback",
                                             200,
                                             &node::on_feedback,
                                             this,
                                             ros::TransportHints().tcpNoDelay() ) )
          , lua_def_( handle_.advertiseService( "/control/lua/def", &node::on_lua_def, this ) )
          , lua_call_( handle_.advertiseService( "/control/lua/call", &node::on_lua_call, this ) )
          , context_comp_( handle_, std::move( rob ), config_ )
          , ctl_comp_( handle_, get_context() )
          , telem_comp_( handle_, config_ )
        {
        }

        bool on_lua_def( schpin_msgs::LuaDef::Request&, schpin_msgs::LuaDef::Response& resp )
        {
                resp.definition = lua_vm_.get_json_def();
                return true;
        }

        bool on_lua_call( schpin_msgs::LuaCall::Request& req, schpin_msgs::LuaCall::Response& resp )
        {
                nlohmann::json j = nlohmann::json::parse( req.args );
                SCHPIN_INFO_LOG( "node", "got call: " << req.name << " res: " << j );
                lua_vm_.exec_bind( req.name, j, get_context(), feedback_handler{ *this } )
                    .match(
                        [&]( std::optional< nlohmann::json > opt_j ) {
                                resp.success = true;
                                if ( opt_j ) {
                                        resp.response = opt_j->dump();
                                }
                        },
                        [&]( auto e ) {
                                resp.success  = false;
                                resp.response = to_string( e );
                        } );
                return true;
        }

        auto comps()
        {
                return std::tie( context_comp_, ctl_comp_, telem_comp_ );
        }

        const node_context& get_context()
        {
                return context_comp_.get_context();
        }

        void force_body_pose( const slua_pose_input_var& ivar )
        {
                pose< world_frame > bpose =
                    slua_get_real_pose( ivar, get_context().rstate.body_pose );
                context_comp_.set_body_pose( bpose );
        }

        void set_servo_config( servo_id id, const nlohmann::json& j )
        {
                context_comp_.set_servo_config( id, j, feedback_handler{ *this } );
        }

        void run()
        {
                SCHPIN_INFO_LOG( "control", "Blob initialized and running" );
                ros::Rate r( 60 );
                while ( ros::ok() ) {
                        tick();
                        ros::spinOnce();
                        r.sleep();
                }
        }

        void tick()
        {
                em::for_each( comps(), [&]( auto& comp ) {
                        comp.tick( get_context(), feedback_handler{ *this } );
                } );
        }

        void on_feedback( const schpin_msgs::BlobCmd::ConstPtr& msg )
        {
                auto opt_msg = fw_node_message::make( msg->data );
                if ( !opt_msg ) {
                        SCHPIN_ERROR_LOG( "node",
                                          "input msg is too long: " << em::view{ msg->data } );
                        return;
                }

                parse_fw_node( *opt_msg )
                    .match(
                        em::uncurry |
                            [&]( ctl_id cid, fw_node_group_variant subvar, checksum ) {
                                    em::for_each( comps(), [&]( auto& comp ) {
                                            comp.on_feedback(
                                                subvar, get_context(), feedback_handler{ *this } );
                                    } );
                                    context_comp_.feedback_processed( cid );
                            },
                        [&]( auto e ) {
                                SCHPIN_ERROR_LOG( "node", e << " for " << *opt_msg );
                        } );
        }

        ctl_id transmit( node_fw_group_message mbm )
        {
                ctl_id msg_id = msg_id_;
                msg_id_++;

                auto m = node_fw_protocol::serialize( msg_id, mbm, checksum{} );

                schpin_msgs::BlobCmd msg;
                msg.data = map_f_to_v( m, []( uint8_t val ) {
                        return val;
                } );
                control_pub_.publish( msg );
                context_comp_.record_msg_sent( msg_id, m );

                return msg_id;
        }
};

int main( int argc, char* argv[] )
{

        ros::init( argc, argv, "node" );
        ros::NodeHandle handle{ "~" };

        ros::Duration( 0.5 ).sleep();

        load_node_data().match(
            em::uncurry |
                [&]( node_hardware_config config, robot_model rob ) {
                        node b{ handle, std::move( config ), std::move( rob ) };
                        ros::Duration( 0.5 ).sleep();
                        b.run();
                },
            [&]( const load_error& e ) {
                    std::cerr << e;
            } );
}
