
#include "node/components/telemetry.h"

using namespace schpin;

telemetry_component::telemetry_component( ros::NodeHandle handle, const node_hardware_config& c )
  : handle_( handle )
  , diag_pub_( handle.advertise< diagnostic_msgs::DiagnosticArray >( c.diag_topic, 100 ) )
  , period_( 1.f / c.diag_freq )
  , last_( ros::Time::now() )
{
}

void telemetry_component::tick( const node_context&, sm_feedback_f msg_cb )
{
        if ( ros::Time::now() < last_ + period_ ) {
                return;
        }

        freq_analyzer_.tick();

        diagnostic_msgs::DiagnosticArray msg;
        msg.status = map_f_to_v(
            msgs_, em::uncurry | [&]( servo_id, diagnostic_msgs::DiagnosticStatus status_msg ) {
                    return status_msg;
            } );
        for ( auto submsg : make_freq_msgs() ) {
                msg.status.push_back( std::move( submsg ) );
        }

        msg.status.push_back( make_buffer_size_msg( "input buffer", input_buffer_size_ ) );
        msg.status.push_back( make_buffer_size_msg( "event buffer", event_buffer_size_ ) );

        diag_pub_.publish( msg );

        roulette_.tick( [&]( auto tag ) {
                for ( auto& pr : msgs_ ) {
                        msg_cb( make_message< decltype( tag )::value >( pr.first ) );
                }
        } );

        last_ = ros::Time::now();
}

void telemetry_component::on_feedback( fw_node_group_variant v,
                                       const node_context&       context,
                                       sm_feedback_f )
{
        freq_analyzer_.make_record( v );

        v << tuple_match(                                                 //
            [&]( cmd_tag< ACTIVATE >, servo_id id, activated_enum en ) {  //
                    if ( en == ACTIVATED || en == WAS_ACTIVATED ) {
                            check_record( id );
                    }
            },
            [&]( cmd_tag< INPUT_BUFFER_SIZE >, uint16_t val ) {  //
                    input_buffer_size_ = val;
            },
            [&]( cmd_tag< EVENT_BUFFER_SIZE >, uint16_t val ) {  //
                    event_buffer_size_ = val;
            },
            [&]( cmd_tag< ERROR >, ctl_id id, error_message e ) {
                    SCHPIN_ERROR_LOG( "telem", "incoming error:" << e );
                    context.hist.try_log( id );
            },
            [&]( cmd_tag< SERVO >, servo_id id, servo_master_command_variant var ) {
                    var << tuple_match( [&]( auto... args ) {  //
                            on_servo_msg( id, context, args... );
                    } );
            },
            [&]( auto... ) {} );
}
void telemetry_component::update_key( servo_id           sid,
                                      const std::string& key,
                                      const std::string& value )
{
        check_record( sid );

        diagnostic_msgs::KeyValue new_pair;
        new_pair.key   = key;
        new_pair.value = value;

        auto iter = em::find_if( msgs_[sid].values,
                                 [&]( const auto& kval ) {  //
                                         return kval.key == key;
                                 } );
        if ( iter == msgs_[sid].values.end() ) {
                msgs_[sid].values.push_back( new_pair );
        } else {
                *iter = new_pair;
        }
}

diagnostic_msgs::DiagnosticStatus& telemetry_component::ref_msg( servo_id sid )
{
        auto iter = msgs_.find( sid );
        if ( iter == msgs_.end() ) {
                auto& msg       = msgs_[sid];
                msg.level       = diagnostic_msgs::DiagnosticStatus::OK;
                msg.name        = "Servomotor/" + std::to_string( sid );
                msg.message     = "Working correctly";
                msg.hardware_id = "SM" + std::to_string( sid );
                return msg;
        }
        return iter->second;
}

std::vector< diagnostic_msgs::DiagnosticStatus > telemetry_component::make_freq_msgs()
{
        return map_f_to_v(
            freq_analyzer_.get_stats(), em::uncurry | [&]( messages_enum msg_id, auto datasets ) {
                    diagnostic_msgs::DiagnosticStatus msg;
                    msg.level = diagnostic_msgs::DiagnosticStatus::OK;
                    msg.name = "Feedback channel/" + std::string{ magic_enum::enum_name( msg_id ) };
                    msg.message     = "Working";
                    msg.hardware_id = "control";

                    for ( auto& dat : datasets ) {
                            diagnostic_msgs::KeyValue pr;
                            pr.key   = "Window " + std::to_string( dat.dur.toSec() );
                            pr.value = std::to_string( dat.freq );
                            msg.values.push_back( pr );
                    }

                    return msg;
            } );
}

diagnostic_msgs::DiagnosticStatus
telemetry_component::make_buffer_size_msg( const std::string& name, uint16_t val )
{
        diagnostic_msgs::DiagnosticStatus msg;
        msg.level       = diagnostic_msgs::DiagnosticStatus::OK;
        msg.name        = "Buffer size/" + name;
        msg.message     = "Working";
        msg.hardware_id = "control";

        diagnostic_msgs::KeyValue pr;
        pr.key   = "size";
        pr.value = std::to_string( val );

        msg.values.push_back( pr );
        return msg;
}
