#include "fw/serial.h"
#include "fw/servo/holder.h"
#include "node/lua_vm/def.h"
#include "node/lua_vm/extra_ser.h"

#include <ros/ros.h>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

using namespace schpin;

// TODO:
//
//    make extension to protocol that allows json -> protocol serialization

struct msg_event
{
        nlohmann::json j;
};

struct msg_bind
{

        static constexpr std::string_view name = "msg";
        static constexpr std::string_view desc = "creates message out of json";

        inline static const auto arg =
            slua_struct< msg_event >( lua_key_arg< nlohmann::json >{ "json", "data" } );
        inline static const no_return_type res{};
};

struct slua_runner_def
{
        static auto get_def()
        {
                return std::make_tuple( slua_callback< msg_bind >{} );
        }
};

template < auto ID >
struct nlohmann::adl_serializer< cmd_tag< ID > >
{
        static void to_json( nlohmann::json& j, cmd_tag< ID > )
        {
                j = magic_enum::enum_name( ID );
        }
        static void from_json( const nlohmann::json& j, cmd_tag< ID >& id )
        {

                auto opt_val = magic_enum::enum_cast< decltype( ID ) >( j.get< std::string >() );
                if ( !opt_val.has_value() ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "Failed to parse ID" );
                }

                if ( opt_val.value() != ID ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "ID does not match" );
                }

                id = cmd_tag< ID >{};
        }
};

template < typename >
struct variant_serializer;

template < typename... Ts >
struct variant_serializer< std::variant< Ts... > >
{
        template < typename T >
        struct tag
        {
                using value_type = T;
        };

        static void to_json( nlohmann::json& j, const std::variant< Ts... >& var )
        {
                std::visit(
                    [&]( const auto& item ) {
                            j = item;
                    },
                    var );
        }

        static std::variant< Ts... > from_json( const nlohmann::json& j )
        {
                std::optional< std::variant< Ts... > > opt_res;

                auto f = [&]( auto tagv ) {
                        using T = typename decltype( tagv )::value_type;
                        try {
                                opt_res = std::variant< Ts... >{ j.get< T >() };
                                return true;
                        }
                        catch ( ... ) {
                                return false;
                        }
                };

                ( f( tag< Ts >{} ) || ... || false );
                if ( !opt_res ) {
                        throw nlohmann::json::parse_error::create(
                            -1, 0, "no variant alternative can be parsed" );
                }
                return *opt_res;
        }
};

template <>
struct nlohmann::adl_serializer< master_servo_command_variant >
  : variant_serializer< master_servo_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< master_servo_config_variant >
  : variant_serializer< master_servo_config_variant >
{
};

template <>
struct nlohmann::adl_serializer< typename prot_config_types< servo_config >::parameters_variant >
  : variant_serializer< typename prot_config_types< servo_config >::parameters_variant >
{
};

class servo_runner
{
        serial_connection          ser_;
        servo_holder               servo_;
        slua_vm< slua_runner_def > vm_;
        slua_repl                  repl_;

      public:
        servo_runner( serial_connection ser )
          : ser_( ser )
          , servo_( test_servo_tag{}, lewan_servo_id::get< 13 >(), get_system_time() )
        {
        }

        // TODO: duplication from control_pc :/
        auto make_matcher()
        {
                return match(
                    [this]( control_tag, lewan_servo_id, auto ) {  //
                    },
                    [this]( activate_tag, servo_id id ) {
                            SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                    },
                    [this]( servo_tag, s_ctl_id msg_id, reply r, const auto& m ) {
                            ser_.write( m );
                            if ( r == reply::NO ) {
                                    return;
                            }
                            receive_serial_msg( msg_id );
                    } );
        }
        void receive_serial_msg( s_ctl_id msg_id )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;
                ser_.read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    parse_message( msg_id, m );
                            } );

                            if ( !load_fired ) {
                                    SCHPIN_ERROR_LOG(
                                        "control",
                                        "servo not replying, got:" << convert_view< int >( data ) );
                            }
                    },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "control", err );
                    } );
        }
        void parse_message( s_ctl_id msg_id, const lewan_master_message& m )
        {
                lewan_master_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m );
                                } )
                    .match(
                        [&]( lewan_master_variant msg_var ) {
                                servo_.on_servo_receive( get_system_time(), msg_id, msg_var );
                        },
                        [&]( auto err ) {
                                SCHPIN_ERROR_LOG( "lewan_servo", err );
                        } );
        }

        void tick()
        {
                servo_.tick( get_system_time(), make_matcher() );
                servo_.poll( get_system_time(), 0, make_matcher() );
                std::string msg = repl_.prompt();
                vm_.load( msg ).optionally_log();
                vm_.full_exec( [&]( lua_State*, auto event ) {
                           handle_event( event );
                   } )
                    .optionally_log();
        }

        void handle_event( msg_event e )
        {
                auto var = e.j.get< master_servo_command_variant >();

                servo_.on_servo_packet( ctl_id{ 0 }, var )
                    .match(
                        [&]( servo_event e ) {
                                servo_.on_event(
                                    ctl_id{ 0 }, get_system_time(), e.sub_event, make_matcher() );
                        },
                        [&]( servo_master_message m ) {
                                // TODO: this should return as json
                                SCHPIN_INFO_LOG( "servo_runner", m );
                        } );
        }
};

struct servo_tester_config
{
        std::filesystem::path dev;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_opt_param< std::filesystem::path >(
                        "device", "Device to open for connection", "/dev/ttyUSB0" ) );
        }
};

int main( int argc, char* argv[] )
{

        ros::init( argc, argv, "servo_tester" );
        ros::AsyncSpinner spinner{ 2 };
        spinner.start();

        get_config< servo_tester_config >( "~" ).match(
            [&]( servo_tester_config conf ) {
                    serial_connection::make( conf.dev )
                        .match(
                            [&]( serial_connection p ) {
                                    servo_runner runnner{ p };

                                    while ( true ) {
                                            runnner.tick();
                                    }
                            },
                            [&]( auto e ) {
                                    std::cerr << e;
                            } );
            },
            [&]( auto e ) {
                    std::cerr << e;
            } );
}

