#include <atomic>
#include <schpin_lib/algorithm.h>
#include <schpin_msgs/ProfilerGroup.h>
#include <schpin_msgs/ProfilerGroupRecord.h>
#include <schpin_msgs/ProfilerRecord.h>

#pragma once

namespace schpin
{

using TimePoint = std::chrono::time_point< std::chrono::high_resolution_clock >;

using ProfilerMsgTuple = std::tuple< uint32_t, uint32_t >;

template < typename Mark >
struct ProfilerGroupBy
{
        Mark m;
};

template < typename Mark >
struct ProfilerRecord
{
        TimePoint from{};
        TimePoint to{};
        Mark      mark;
};

template < typename, bool, typename... >
class LocalTimer;

template < typename Mark, typename... Settings >
class ProfilerAnalyzer
{
        std::vector< ProfilerRecord< Mark > >     buffer_;
        std::size_t                               i_;
        ros::NodeHandle                           handle_;
        std::map< Mark, ros::Publisher >          pubs_;
        std::map< Mark, ros::Publisher >          group_pubs_;
        std::tuple< std::decay_t< Settings >... > settings_;

      public:
        ProfilerAnalyzer( std::string name, Settings... s )
          : buffer_()
          , i_( 0 )
          , handle_( "~/" + name )
          , settings_( s... )
        {
        }

        void insert( ProfilerRecord< Mark > record )
        {
                if ( buffer_.size() == static_cast< std::size_t >( std::pow( 2, 16 ) ) ) {
                        buffer_[i_] = record;
                        i_          = ( i_ + 1 ) % buffer_.size();
                } else {
                        buffer_.push_back( record );
                }
        }

        struct GroupItem
        {
                uint32_t                        count;
                std::chrono::duration< double > sum_time;
        };
        struct GroupRecord
        {
                TimePoint                   from;
                TimePoint                   to;
                std::map< Mark, GroupItem > items{};
        };

        schpin_msgs::ProfilerGroupRecord
        calculate_group_record( Mark m, const std::vector< GroupRecord >& group_records )
        {
                schpin_msgs::ProfilerGroupRecord gr;
                gr.name                  = to_string( m );
                gr.min_count             = std::numeric_limits< uint32_t >::max();
                gr.min_sum_time          = std::numeric_limits< float >::max();
                std::size_t record_count = 0;
                for ( const GroupRecord& r : group_records ) {
                        auto iter = r.items.find( m );
                        if ( iter == r.items.end() ) {
                                continue;
                        }
                        record_count += 1;

                        uint32_t count = iter->second.count;
                        gr.count += count;
                        gr.min_count = std::min( gr.min_count, count );
                        gr.max_count = std::max( gr.max_count, count );

                        float t = static_cast< float >( iter->second.sum_time.count() );
                        gr.sum_time += t;
                        gr.min_sum_time = std::min( gr.min_sum_time, t );
                        gr.max_sum_time = std::max( gr.max_sum_time, t );
                }

                gr.avg_count    = float( gr.count ) / static_cast< float >( record_count );
                gr.avg_sum_time = gr.sum_time / static_cast< float >( record_count );

                return gr;
        }

        void handle( ProfilerGroupBy< Mark > group_by )
        {

                Mark m = group_by.m;

                std::vector< GroupRecord > records;
                for ( const ProfilerRecord< Mark >& r : buffer_ ) {
                        if ( r.mark == m ) {
                                records.push_back( GroupRecord{ r.from, r.to } );
                        }
                }

                std::set< Mark > seen_keys;

                for ( const ProfilerRecord< Mark >& r : buffer_ ) {
                        auto iter = em::find_if( records, [&]( const GroupRecord& record ) {
                                return record.from <= r.from && r.to <= record.to;
                        } );
                        if ( iter == records.end() ) {
                                return;
                        }
                        GroupItem& item = iter->items[r.mark];
                        item.count += 1;
                        item.sum_time += r.to - r.from;
                        seen_keys.insert( r.mark );
                }

                schpin_msgs::ProfilerGroup main_msg;
                main_msg.name     = to_string( m );
                main_msg.subcalls = map_f_to_v(
                    seen_keys, [&]( Mark k ) { return calculate_group_record( k, records ); } );

                auto iter = group_pubs_.find( m );
                if ( iter == group_pubs_.end() ) {
                        group_pubs_[m] = handle_.advertise< schpin_msgs::ProfilerGroup >(
                            to_string( m ) + "/groups", 100 );
                }
                group_pubs_[m].publish( main_msg );
        }

        void makeReport()
        {
                if ( buffer_.empty() ) {
                        return;
                }

                std::map< Mark, std::vector< float > > data;
                for ( const ProfilerRecord< Mark >& r : buffer_ ) {
                        std::chrono::duration< float > diff = r.to - r.from;
                        data[r.mark].push_back( diff.count() );
                }
                for ( auto [m, timestamps] : data ) {
                        if ( timestamps.empty() ) {
                                continue;
                        }
                        schpin_msgs::ProfilerRecord record;

                        record.name   = to_string( m );
                        record.median = timestamps[timestamps.size() / 2];
                        record.avg    = em::avg( timestamps );
                        record.count  = static_cast< unsigned >( timestamps.size() );
                        record.max    = 0;
                        record.min    = std::numeric_limits< float >::max();

                        for ( float f : timestamps ) {
                                record.max = std::max( record.max, f );
                                record.min = std::min( record.min, f );
                                record.variance =
                                    static_cast< float >( std::pow( f - record.avg, 2 ) );
                        }

                        auto iter = pubs_.find( m );
                        if ( iter == pubs_.end() ) {
                                pubs_[m] = handle_.advertise< schpin_msgs::ProfilerRecord >(  //
                                    to_string( m ) + "/records",
                                    100 );
                        }
                        pubs_[m].publish( record );
                }
                em::for_each( settings_, [&]( auto s ) { handle( s ); } );
        }
};

template < typename Mark, bool enable_export = true, typename... Settings >
class Profiler
{
        std::string                                            name_;
        std::vector< ProfilerRecord< Mark > >                  records_;
        std::atomic< std::uint16_t >                           i_      = 0;
        std::uint16_t                                          last_i_ = 0;
        std::optional< ProfilerAnalyzer< Mark, Settings... > > analyzer_;
        std::optional< ros::Timer >                            timer_;

        // WARING: so, this uses i_ to idnex circular buffer, normaly you increment like: i_ = (i_ +
        // 1) % m however that is tricky to do threadsafe, so instead i_ of size uint16_t is used,
        // buffer has size 2^16, and instead of % operation, unsigned overflow is used.

      public:
        Profiler( std::string name, Settings... s )
          : name_( name )
          , records_( std::pow( 2, 16 ) )
        {
                if constexpr ( enable_export ) {
                        analyzer_ = std::make_optional< ProfilerAnalyzer< Mark, Settings... > >(
                            name, s... );
                        timer_ = ros::NodeHandle{ "~" }.createTimer(
                            ros::Duration( 0.1 ), [&]( const auto& event ) { onTimer( event ); } );
                }
        }

        LocalTimer< Mark, enable_export, Settings... > make_local_timer( Mark mark )
        {
                return LocalTimer< Mark, enable_export, Settings... >( this, mark );
        }

        void pushRecord( ProfilerRecord< Mark > record )
        {
                records_[i_++] = record;
        }

        void onTimer( const ros::TimerEvent& )
        {
                while ( last_i_ != i_ ) {
                        if constexpr ( enable_export ) {
                                analyzer_->insert( records_[last_i_++] );
                        } else {
                                return;
                        }
                }
                analyzer_->makeReport();
        }
};

template < typename Mark, bool enable_export, typename... Settings >
class LocalTimer
{

        Profiler< Mark, enable_export, Settings... >* profiler_;
        ProfilerRecord< Mark >                        record_;

      public:
        LocalTimer( Profiler< Mark, enable_export, Settings... >* profiler, Mark mark )
          : profiler_( profiler )
        {
                record_.mark = mark;
                record_.from = std::chrono::high_resolution_clock::now();
        }

        LocalTimer( const LocalTimer& ) = delete;
        LocalTimer& operator=( const LocalTimer& ) = delete;

        ~LocalTimer()
        {
                record_.to = std::chrono::high_resolution_clock::now();
                profiler_->pushRecord( record_ );
        }
};

}  // namespace schpin
