#include "magic_enum.hpp"
#include "schpin_lib_exp/convert_iterator.h"
#include "schpin_lib_exp/overload.h"
#include "schpin_lib_exp/protocol/base.h"
#include "schpin_lib_exp/protocol/item.h"
#include "schpin_lib_exp/protocol/tuple.h"

#pragma once

namespace schpin
{

template < auto ID, typename... Args >
struct command
{
        using command_id = decltype( ID );
        using prot_tuple = protocol_tuple<>::with_items< command_id, Args... >;
        using value_tuple =
            std::tuple< cmd_tag< ID >, typename protocol_item< Args >::value_type... >;

        static constexpr std::size_t arg_n = prot_tuple::item_n - 1;
        static constexpr std::size_t size  = prot_tuple::size;
        static constexpr command_id  id    = ID;
        static constexpr auto        hash =
            serialize( ID, sizeof...( Args ), protocol_item< Args >::hash... );

        template < typename... NewArgs >
        using with_args = command< ID, Args..., NewArgs... >;

        template < typename... NewArgs >
        using with_promoted_args =
            command< ID, Args..., protocol_recursive_promoter< NewArgs >... >;
};

// TODO: this should get proper place
namespace detail::stringify
{
        template < typename... Args >
        inline std::ostream& operator<<( std::ostream& os, const std::variant< Args... >& var );

        template < typename... Args >
        inline std::ostream& operator<<( std::ostream& os, const std::tuple< Args... >& var )
        {
                os << "{";
                char del = ' ';
                em::for_each( var, [&]( const auto& item ) {
                        os << del << item;
                        del = ',';
                } );
                os << " }";
                return os;
        }

        template < typename... Args >
        inline std::ostream& operator<<( std::ostream& os, const std::variant< Args... >& var )
        {
                os << "[ ";
                var << match( [&]( const auto& item ) {
                        os << item;
                } );
                os << "]";
                return os;
        }
}  // namespace detail::stringify

template < typename MsgCounterType = uint8_t, typename... Commands >
class command_group
{
        struct no_cmd
        {
                using command_id = uint8_t;
        };
        const std::tuple< Commands... > commands_;

      public:
        template < typename NewMsgCounterType >
        using with_msg_counter = command_group< NewMsgCounterType, Commands... >;

        template < typename... NewCommands >
        using with_commands = command_group< MsgCounterType, Commands..., NewCommands... >;

        static constexpr std::size_t commands_n = sizeof...( Commands );
        static constexpr std::size_t max_size = std::max( { std::size_t{ 0 }, Commands::size... } );
        static constexpr auto        commands_hashes = std::make_tuple( Commands::hash... );

        static_assert( ( ... && ( sizeof( typename Commands::command_id ) == 1 ) ),
                       "ID for commands have to be of size 1" );
        // static_assert( protocol::hashesh_are_unique< commands_n >{{Commands::hash...}} );

        using message_t       = prot_message< MsgCounterType, max_size, command_group >;
        using command_variant = std::variant< typename Commands::value_tuple... >;
        using value_type      = command_variant;
        using command_id =
            typename std::tuple_element_t< 0, std::tuple< Commands..., no_cmd > >::command_id;

        static std::string stringify( command_variant v )
        {
                using namespace detail::stringify;

                std::stringstream ss;
                ss << v;
                return ss.str();
        }

        static std::string stringify( const message_t& m )
        {
                std::stringstream ss;

                extract( m ).match(
                    [&]( command_variant v ) {
                            ss << " [ " << stringify( std::move( v ) )
                               << " :: " << convert_view< int >( m ) << " ] ";
                    },
                    [&]( auto ) {
                            ss << em::view{ m };
                    } );
                return ss.str();
        }

        template < typename Message >
        static em::either< command_variant, protocol_error_record > extract( Message&& message )
        {
                em::either< command_variant, protocol_error_record > res = protocol_error_record{};
                match(
                    message,
                    [&]( error_tag, protocol_error_record rec ) {
                            res = rec;
                    },
                    [&]( auto cmd, auto... items ) {
                            res = command_variant{ std::make_tuple( cmd, items... ) };
                    } );

                return res;
        }

        template < typename Message, typename... Visitors >
        static constexpr std::size_t match( Message&& message, Visitors&&... vis )
        {
                return match( message, schpin::match( vis... ) );
        }

        template < typename Message, typename Visitor >
        static constexpr std::size_t match( Message&& message, Visitor&& vis )
        {
                std::size_t used = 0;
                auto        f    = [&]( auto cmd ) -> bool {
                        using command    = decltype( cmd );
                        using command_id = typename command::command_id;
                        if ( command::id != message[0] ) {
                                return false;
                        }
                        used = command::prot_tuple::match(
                            message,
                            [&]( error_tag, auto err ) {
                                    vis( error_tag{}, err );
                            },
                            [&]( command_id, auto... values ) {
                                    vis( cmd_tag< command::id >{}, values... );
                            } );
                        return true;
                };

                bool matched = ( f( Commands{} ) || ... );
                if ( !matched ) {
                        vis( error_tag{}, protocol_error_record{ MISSING_MATCH_E, 0 } );
                }

                return used;
        }

        template < auto ID, typename... Args >
        static auto make_val( Args... args )
        {
                // TODO: duplicated from below
                constexpr bool id_exists = ( ( Commands::id == ID ) || ... );

                static_assert( id_exists,
                               "you cannot create this message, the id is not present in command "
                               "choice" );

                std::optional< command_variant > opt_res;
                auto                             f = [&]( auto cmd ) {
                        using command = decltype( cmd );
                        if constexpr ( command::id == ID ) {
                                opt_res =
                                    command_variant{ command::prot_tuple::make_val( ID, args... ) };
                        }
                };

                ( f( Commands{} ), ... );
                SCHPIN_ASSERT( opt_res );
                return *opt_res;
        }

        template < auto ID, typename... Args >
        static constexpr message_t create( Args... args )
        {
                constexpr bool id_exists = ( ( Commands::id == ID ) || ... );

                static_assert( id_exists,
                               "you cannot create this message, the id is not present in command "
                               "choice" );

                message_t res;
                auto      f = [&]( auto cmd ) {
                        using command = decltype( cmd );
                        if constexpr ( command::id == ID ) {
                                static_assert( std::is_same_v< decltype( ID ),
                                                               typename command::command_id > );
                                static_assert( command::arg_n == sizeof...( Args ),
                                               "Count of arguments for command has to be same as "
                                               "number of arguments defined for it." );
                                res = message_t{ command::prot_tuple::serialize( ID, args... ) };
                        }
                };
                ( f( Commands{} ), ... );
                return res;
        }
};

}  // namespace schpin
