#include "schpin_lib_exp/bounded.h"
#include "schpin_lib_exp/protocol/base.h"
#include "schpin_lib_exp/protocol/message.h"
#include "schpin_lib_exp/protocol/sizeless_message.h"
#include "schpin_lib_exp/serialize.h"

#include <bitset>
#include <variant>

#pragma once

namespace schpin
{

template < typename T >
struct protocol_item_base
{
        static std::string stringify( const T& item )
        {
                std::stringstream ss;
                ss << item;
                return ss.str();
        }
};

template < typename T, typename Enable = void >
struct protocol_item;

template < typename T >
struct protocol_item<
    T,
    typename std::enable_if_t< std::is_arithmetic_v< T > || std::is_enum_v< T > > >
  : protocol_item_base< T >
{
        using value_type                  = T;
        static constexpr auto        hash = serialize( protocol_type_id_v< T > );
        static constexpr std::size_t size = serializer< T >::size;

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                                   const T&                      item )
        {
                serializer< T >::template serialize< serialization_endianess< value_type > >(
                    buffer, item );
                return size;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( opt_view ) {
                        visitor( value_tag{},
                                 serializer< T >::template deserialize<
                                     serialization_endianess< value_type > >( *opt_view ) );
                } else {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                }
                return size;
        }
};

template < std::size_t N >
struct protocol_item< std::bitset< N > > : protocol_item_base< std::bitset< N > >
{
        using value_type                  = std::bitset< N >;
        static constexpr auto        hash = serialize( protocol_type_id_v< value_type >, N );
        static constexpr std::size_t size = ( N + 7 ) / 8;

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                                   const std::bitset< N >&       item )
        {
                for ( std::size_t i : em::range( size ) ) {
                        std::bitset< 8 > byte;
                        for ( std::size_t j : em::range< std::size_t >( 8u ) ) {
                                byte[j] = item[i * 8 + j];
                        }
                        buffer[i] = static_cast< uint8_t >( byte.to_ulong() );
                }
                return size;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                std::bitset< N > res;
                for ( std::size_t i : em::range( size ) ) {
                        std::bitset< 8 > byte = buffer[i];
                        for ( std::size_t j : em::range< std::size_t >( 8u ) ) {
                                res[i * 8 + j] = byte[j];
                        }
                }
                visitor( value_tag{}, res );
                return size;
        }
};

template < typename MsgCounterType, std::size_t N, typename Owner >
struct protocol_item< prot_message< MsgCounterType, N, Owner > >
  : protocol_item_base< prot_message< MsgCounterType, N, Owner > >
{
        using value_type   = prot_message< MsgCounterType, N, Owner >;
        using sub_protocol = protocol_item< std::array< uint8_t, N > >;
        using id_protocol  = protocol_item< MsgCounterType >;
        static constexpr auto hash =
            serialize( protocol_type_id_v< value_type >, sub_protocol::hash );
        static constexpr std::size_t size = sub_protocol::size + id_protocol::size;

        static_assert( N < std::numeric_limits< MsgCounterType >::max() );

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* >                   buffer,
                      const prot_message< MsgCounterType, N, Owner >& m )
        {
                std::size_t used =
                    id_protocol::serialize_at( buffer, static_cast< MsgCounterType >( m.size() ) );

                for ( std::size_t i : em::range( m.size() ) ) {
                        buffer[i + used] = m[i];
                }
                return m.size() + used;
        }
        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                if ( buffer.empty() ) {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                        return 0;
                }

                std::optional< MsgCounterType > opt_real_size;
                // TOOD: prime example taht deserialize API is not good
                std::size_t used =
                    id_protocol::deserialize( buffer,
                                              match(
                                                  [&]( value_tag, MsgCounterType val ) {
                                                          opt_real_size = val;
                                                  },
                                                  [&]( error_tag, protocol_error_record rec ) {
                                                          visitor( error_tag{}, rec );
                                                  } ) );

                if ( !opt_real_size ) {
                        return 0;
                }
                MsgCounterType real_size = *opt_real_size;

                if ( buffer.size() - used < real_size ) {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                        return 0;
                }
                std::array< uint8_t, N > data;
                for ( std::size_t i = 0; i < real_size; ++i ) {
                        data[i] = buffer[i + used];
                }

                visitor( value_tag{}, value_type{ data, real_size } );
                return static_cast< std::size_t >( real_size ) + used;
        }
};

template < typename... Ts >
struct protocol_item< std::variant< Ts... > > : protocol_item_base< std::variant< Ts... > >
{
        using id_type = uint8_t;
        static_assert( sizeof...( Ts ) < std::numeric_limits< id_type >::max() );

        using value_type = std::variant< Ts... >;
        static constexpr std::size_t size =
            sizeof( id_type ) + std::max( { protocol_item< Ts >::size... } );

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                                   const value_type&             var )
        {
                buffer[0]     = static_cast< id_type >( var.index() );
                auto sub_buff = *try_static_view< size - 1 >( buffer.begin() + 1, buffer.end() );
                std::size_t sub_used =
                    var << match( [&]( const auto& item ) {
                            return protocol_item< std::decay_t< decltype( item ) > >::serialize_at(
                                sub_buff, item );
                    } );
                return sub_used + 1;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& vis )
        {
                std::size_t used  = 0;
                bool        fired = false;
                em::for_each(
                    em::make_sequence_tuple_t< sizeof...( Ts ) >{}, [&]( auto integ_const ) {
                            using I = decltype( integ_const );
                            using T = std::variant_alternative_t< I::value, value_type >;
                            if ( buffer[0] != I::value ) {
                                    return;
                            }
                            fired = true;
                            used  = protocol_item< T >::deserialize( em::tail( buffer ), vis );
                    } );
                if ( !fired ) {
                        vis( error_tag{}, protocol_error_record{ BAD_ARGUMENT_E, 0 } );
                }
                return used;
        }
};

template < std::size_t N >
struct protocol_item< prot_sizeless_message< N > >
  : protocol_item_base< prot_sizeless_message< N > >
{
        using value_type                  = prot_sizeless_message< N >;
        static constexpr std::size_t size = N;

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* >     buffer,
                                                   const prot_sizeless_message< N >& m )
        {
                for ( std::size_t i : em::range( m.size() ) ) {
                        buffer[i] = m[i];
                }
                return m.size();
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                if ( buffer.size() > N ) {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                        return 0;
                }

                em::static_vector< uint8_t, N > res;
                for ( uint8_t val : buffer ) {
                        res.push_back( val );
                }

                visitor( value_tag{}, prot_sizeless_message< N >{ res } );

                return res.size();
        }
};

template < typename T >
struct protocol_recursive_promoter;

template < typename MsgCounterType, std::size_t N, typename Owner >
struct protocol_item< protocol_recursive_promoter< prot_message< MsgCounterType, N, Owner > > >
{
        using sub_protocol = protocol_item< prot_message< MsgCounterType, N, Owner > >;

        using value_type = typename Owner::value_type;

        static constexpr auto        hash = sub_protocol::hash;
        static constexpr std::size_t size = sub_protocol::size;

        static std::string stringify( const value_type& item )
        {
                return Owner::stringify( item );
        }

        static constexpr std::size_t
        serialize_at( static_view< size, uint8_t* >                   buffer,
                      const prot_message< MsgCounterType, N, Owner >& n )
        {
                return sub_protocol::serialize_at( buffer, n );
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                return sub_protocol::deserialize(  //
                    buffer,
                    match(
                        [&]( value_tag, prot_message< MsgCounterType, N, Owner > msg ) {
                                Owner::extract( msg ).match(
                                    [&]( value_type v ) {
                                            visitor( value_tag{}, v );
                                    },
                                    [&]( auto e ) {
                                            visitor( error_tag{}, e );
                                    } );
                        },
                        [&]( error_tag, auto e ) {
                                visitor( error_tag{}, e );
                        } ) );
        }
};

template < typename T >
struct protocol_item<
    T,
    typename std::enable_if_t< std::is_base_of_v< em::quantity< T, typename T::value_type >, T > > >
  : protocol_item_base< T >
{
        using value_type = T;
        using inner_type = typename T::value_type;
        static_assert( std::is_trivial_v< inner_type > );
        static constexpr auto hash =
            serialize( protocol_type_id_v< inner_type > );  // TODO: add T here into hash
        static constexpr std::size_t size = serializer< inner_type >::size;

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                                   const T&                      item )
        {
                serializer< inner_type >::template serialize<
                    serialization_endianess< value_type > >( buffer, *item );
                return size;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                        return 0;
                }
                inner_type item = serializer< inner_type >::template deserialize<
                    serialization_endianess< value_type > >( *opt_view );
                visitor( value_tag{}, T{ item } );
                return size;
        }
};

template < typename T, T Min, T Max >
struct protocol_item< bounded< T, Min, Max > >
{
        using value_type = bounded< T, Min, Max >;
        static_assert( std::is_trivial_v< T > );
        static constexpr auto hash =
            serialize( protocol_type_id_v< value_type >, protocol_type_id_v< T >, Min, Max );
        static constexpr std::size_t size = serializer< T >::size;

        static std::string stringify( const value_type& item )
        {
                std::stringstream ss;
                ss << std::to_string( *item );
                return ss.str();
        }

        static std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                         const bounded< T, Min, Max >& item )
        {
                serializer< T >::template serialize< serialization_endianess< value_type > >(
                    buffer, *item );
                return size;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                auto opt_view = try_static_view< size >( buffer );
                if ( !opt_view ) {
                        visitor( error_tag{}, protocol_error_record{ INVALID_SIZE_E, 0 } );
                        return 0;
                }
                T item =
                    serializer< T >::template deserialize< serialization_endianess< value_type > >(
                        *opt_view );
                auto opt_range = bounded< T, Min, Max >::make( item );
                if ( !opt_range ) {
                        visitor( error_tag{}, protocol_error_record{ BAD_ARGUMENT_E, 0 } );
                }
                visitor( value_tag{}, *opt_range );
                return size;
        }
};

template < typename T, std::size_t N >
struct protocol_item< std::array< T, N > >
{
        static_assert( N < std::numeric_limits< std::size_t >::max() );

        using value_type                  = std::array< T, N >;
        static constexpr auto        hash = serialize( protocol_type_id_v< value_type >,
                                                static_cast< uint16_t >( N ),
                                                protocol_item< T >::hash );
        static constexpr std::size_t size = protocol_item< T >::size * N;

        static std::string stringify( const std::array< T, N >& data )
        {
                std::stringstream ss;
                for ( const T& item : data ) {
                        ss << item << "  ";
                }
                return ss.str();
        }

        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                                   const std::array< T, N >&     item )
        {
                auto buffers = buffer.template split< N >();
                for ( std::size_t i = 0; i < N; ++i ) {
                        protocol_item< T >::serialize_at( buffers[i], item[i] );
                }
                return size;
        }

        template < typename Buffer, typename Visitor >
        static std::size_t deserialize( Buffer&& buffer, Visitor&& visitor )
        {
                std::array< T, N > res;
                bool               is_ok = true;
                auto               beg   = buffer.begin();
                for ( std::size_t i = 0; i < N; i++ ) {
                        em::view actual_buff{ beg, buffer.end() };
                        beg += protocol_item< T >::deserialize(  //
                            actual_buff,
                            match(
                                [&]( value_tag, T value ) {
                                        res[i] = value;
                                },
                                [&]( error_tag, auto e ) {
                                        e.byte_index +=
                                            static_cast< std::size_t >( beg - buffer.begin() );
                                        visitor( error_tag{}, e );
                                        is_ok = false;
                                } ) );
                }

                if ( is_ok ) {
                        visitor( value_tag{}, res );
                }

                return size;
        }
};

}  // namespace schpin
