#include "body.h"

#include "octree/insert.h"

namespace schpin
{

oc_tree< occupancy_content, robot_frame >
create_body_occupancy_tree( const std::vector< triangle< 3, robot_frame > >& surface,
                            em::length                                       edge_length )
{
        oc_tree< occupancy_content, robot_frame > res( occupancy_content( oc_min_occupancy ),
                                                       edge_length );

        using checker = fast_collision_checker< triangle_collision_query< 3, robot_frame >,
                                                triangle_fast_collision_query< robot_frame >,
                                                occupancy_content >;
        using visitor = batch_collision_visitor< checker, 32 >;

        for ( const triangle< 3, robot_frame >& t : surface ) {
                res.expand( t );
        }

        for ( const auto& batch : make_batches< 32 >( surface ) ) {
                visitor vis{ batch, occupancy_token::FULL };
                res.modify( vis.get_basic_selection(), vis );
        }

        return res;
}
em::either< body, body_boot_error > load_body( const urdf::Model& model,
                                               const std::string& link_name )
{
        urdf::LinkConstSharedPtr link = model.getLink( link_name );

        if ( !link ) {
                return { E( body_boot_error ) << "Failed to load the body link;" };
        }

        return extract_model< robot_frame >( *link )
            .convert_left( [&]( structure_model< robot_frame > smodel ) {
                    auto bound_cube = dimensional_min_max_elem< point< 3, robot_frame > >(
                        em::range( smodel.get_surface().size() * 3 ),
                        [&]( std::size_t i ) {  //
                                return smodel.get_surface()[i / 3][i % 3];
                        } );
                    SCHPIN_INFO_LOG( "body", "Body surface bounding cube: " << bound_cube );
                    return body( std::move( smodel ), em::length( 0.05f ) );
            } )
            .convert_right( [&]( const urdf_parse_error& error ) {
                    return E( body_boot_error ).attach( error ) << "Body not loaded due to error "
                                                                   "in urdf "
                                                                   "parsing";
            } );
}
}  // namespace schpin
