#include "gait/vertex.h"
#include "gait/view.h"

#include <magic_enum.hpp>
#include <pqxx/pqxx>

#pragma once

namespace pqxx
{
template < typename T >
struct string_traits< std::vector< T > >
{
        static bool is_null( const std::vector< T >& vec )
        {
                return vec.empty();
        }

        static std::string to_string( const std::vector< T >& vec )
        {
                std::stringstream ss;
                char              delim = '{';
                for ( const T& val : vec ) {
                        ss << delim << pqxx::string_traits< T >::to_string( val );
                        delim = ',';
                }
                ss << "}";
                return ss.str();
        }
};
}  // namespace pqxx

namespace schpin
{

class gait_db
{

        std::shared_ptr< pqxx::connection > c_ptr;
        long int                            gait_run_id;

      public:
        gait_db()
        {
                try {
                        c_ptr = std::make_shared< pqxx::connection >( "postgresql://"
                                                                      "schpin:schpin@lavinia.fi."
                                                                      "muni.cz/schpin" );
                }
                catch ( ... ) {
                        return;
                }

                c_ptr->prepare( "insert_gait_run",
                                "INSERT INTO gait_run (id, creation, body_path_size, "
                                "sidepath_sizes) "
                                "VALUES ($1,now(),$2,$3)" );
                c_ptr->prepare( "insert_body_path",
                                "INSERT INTO body_path_pose (gait_run_id, pose_i, coord, "
                                "orientation) "
                                "VALUES ($1,$2,$3,$4)" );
                c_ptr->prepare( "insert_sidepath",
                                "INSERT INTO sidepath (gait_run_id, point_i, leg_id, coord) VALUES "
                                "($1,$2,$3,$4)" );
                c_ptr->prepare( "insert_gait_vertex",
                                "INSERT INTO gait_vertex (vertex_id, status, gait_run_id, "
                                "body_path_id, "
                                "sidepath_indexes, leg_states, is_moving, creation_time ) VALUES "
                                "($1, $2, $3, $4, $5, $6, $7, now())" );
                c_ptr->prepare( "insert_gait_edge",
                                "INSERT INTO gait_edge (gait_run_id, source, target, movement_time)"
                                "VALUES"
                                "($1,$2,$3,$4)" );
        }

        void store_gait_run( const gait_body_path&            bpath,
                             const leg_container< sidepath >& sidepaths )
        {
                if ( !c_ptr ) {
                        return;
                }
                pqxx::work tx( *c_ptr );

                auto now    = std::chrono::system_clock::now().time_since_epoch().count();
                gait_run_id = static_cast< int >( now % static_cast< int >( std::pow( 2, 24 ) ) );

                tx.exec_prepared(
                    "insert_gait_run",
                    gait_run_id,
                    bpath.size(),
                    map_f_to_v( sidepaths, [&]( const sidepath& s ) { return s.size(); } ) );

                pqxx::stream_to bp_stream{
                    tx,
                    "body_path_pose",
                    std::vector< std::string >{ "pose_i", "gait_run_id", "coord", "orientation" } };

                for ( auto [id, bp] : em::enumerate( bpath ) ) {
                        bp_stream << std::make_tuple( id,
                                                      gait_run_id,
                                                      map_f_to_v( bp.pos.position ),
                                                      map_f_to_v( bp.pos.orientation ) );
                }
                bp_stream.complete();

                pqxx::stream_to sd_stream{
                    tx,
                    "sidepath",
                    std::vector< std::string >{ "point_i", "gait_run_id", "leg_id", "coord" } };

                for ( auto [leg_id, sidepath] : em::enumerate( sidepaths ) ) {
                        for ( auto [id, p] : em::enumerate( sidepath ) ) {
                                sd_stream
                                    << std::make_tuple( id, gait_run_id, leg_id, map_f_to_v( p ) );
                        }
                }
                sd_stream.complete();

                tx.commit();
        }

        void store_vertex_creation( gait_vertex_id          id,
                                    const gait_state_view&  gview,
                                    gait_vertex_status_enum status )
        {
                if ( !c_ptr ) {
                        return;
                }
                pqxx::work tx( *c_ptr );

                tx.exec_prepared(
                    "insert_gait_vertex",
                    *id,
                    std::string{ magic_enum::enum_name( status ) },
                    gait_run_id,
                    gview.get_state().body_pose_i,
                    map_f_to_v( gview.get_legs(),
                                [&]( const auto& l ) { return l.get_sidepath_index(); } ),
                    map_f_to_v( gview.get_legs(),
                                [&]( const auto& l ) {
                                        leg_vertex_id lid = l.get_state().leg_config;
                                        return map_f_to_v(
                                            l.get_leg_view().get_configuration( lid ),
                                            [&]( em::angle a ) { return *a; } );
                                } ),
                    map_f_to_v( gview.get_legs(),
                                [&]( const auto& l ) { return l.is_leg_moving(); } ) );

                tx.commit();
        }

        void store_vertex_edges( const std::vector< gait_edge >& edges )
        {
                if ( !c_ptr ) {
                        return;
                }
                pqxx::work tx( *c_ptr );

                for ( const gait_edge& edge : edges ) {
                        tx.exec_prepared( "insert_gait_edge",
                                          gait_run_id,
                                          *edge.get_source_id(),
                                          *edge.get_target_id(),
                                          *edge->get_price() );
                }

                tx.commit();
        }
};

}  // namespace schpin
